@@ -51,6 +51,7 @@ typedef enum {
 typedef struct {
   guint64 fd;
   afc_client_t afc_cli;
+  char *app;
 } FileHandle;
 
 typedef struct {
@@ -58,6 +59,7 @@ typedef struct {
   char *id;
   char *icon_path;
   house_arrest_client_t house_arrest;
+  guint num_users;
   afc_client_t afc_cli;
 } AppInfo;
 
@@ -157,7 +159,7 @@ g_vfs_backend_afc_close_connection (GVfsBackendAfc *self)
         {
           afc_client_free (self->afc_cli);
         }
-      else
+      else if (self->mode == ACCESS_MODE_HOUSE_ARREST)
         {
           if (self->apps != NULL)
             {
@@ -176,6 +178,10 @@ g_vfs_backend_afc_close_connection (GVfsBackendAfc *self)
             }
           g_mutex_clear (&self->apps_lock);
         }
+      else
+        {
+          g_assert_not_reached ();
+        }
       g_free (self->model);
       self->model = NULL;
       idevice_free (self->dev);
@@ -261,11 +267,16 @@ g_vfs_backend_sbs_check (sbservices_error_t cond, GVfsJob *job)
 }
 
 static int
-g_vfs_backend_lockdownd_check (lockdownd_error_t cond, GVfsJob *job)
+g_vfs_backend_lockdownd_check (lockdownd_error_t  cond,
+                               GVfsJob           *job,
+                               const char        *internal_job)
 {
   if (G_LIKELY(cond == LOCKDOWN_E_SUCCESS))
         return 0;
 
+  g_debug ("Got lockdown error '%d' while doing '%s'\n",
+           cond, internal_job);
+
   switch (cond)
     {
     case LOCKDOWN_E_INVALID_ARG:
@@ -274,12 +285,20 @@ g_vfs_backend_lockdownd_check (lockdownd_error_t cond, GVfsJob *job)
       break;
     case LOCKDOWN_E_PASSWORD_PROTECTED:
       g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
-                        _("Permission denied"));
+                        _("The device is password protected"));
       break;
     case LOCKDOWN_E_SSL_ERROR:
       g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_CONNECTION_REFUSED,
                         _("Unable to connect"));
       break;
+    case LOCKDOWN_E_USER_DENIED_PAIRING:
+      g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_CONNECTION_REFUSED,
+                        _("User refused to trust this computer"));
+      break;
+    case LOCKDOWN_E_PAIRING_DIALOG_RESPONSE_PENDING:
+      g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_CONNECTION_REFUSED,
+                        _("The user has not trusted this computer"));
+      break;
     default:
       g_vfs_job_failed (job, G_IO_ERROR, G_IO_ERROR_FAILED,
                         _("Unhandled Lockdown error (%d)"), cond);
@@ -428,7 +447,6 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
   afc_error_t aerr;
   const gchar *choices[] = {_("Try again"), _("Cancel"), NULL}; /* keep in sync with the enum above */
   gboolean aborted = FALSE;
-  gchar *message = NULL;
   gint choice;
   gboolean ret;
 
@@ -518,7 +536,7 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
 
   /* first, connect without handshake to get preliminary information */
   lerr = lockdownd_client_new (self->dev, &lockdown_cli, "gvfsd-afc");
-  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "new client, no handshake")))
     goto out_destroy_dev;
 
   /* try to use pretty device name */
@@ -553,7 +571,7 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
   /* set correct freedesktop icon spec name depending on device model */
   value = NULL;
   lerr = lockdownd_get_value (lockdown_cli, NULL, "DeviceClass", &value);
-  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "getting device class")))
     goto out_destroy_lockdown;
 
   plist_get_string_val (value, &self->model);
@@ -580,6 +598,9 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
   /* now, try to connect with handshake */
   retries = 0;
   do {
+    char *message;
+
+    g_debug ("Lockdown client try #%d\n", retries);
     lerr = lockdownd_client_new_with_handshake (self->dev,
                                                 &lockdown_cli,
                                                 "gvfsd-afc");
@@ -589,21 +610,49 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
         continue;
       }
 
-    if (lerr != LOCKDOWN_E_PASSWORD_PROTECTED)
+    if (lerr == LOCKDOWN_E_USER_DENIED_PAIRING)
+      {
+        aborted = TRUE;
+        break;
+      }
+
+    /* An unknown error? Let's try again without prompting */
+    if (lerr == LOCKDOWN_E_UNKNOWN_ERROR)
+      {
+        g_debug ("Got an unknown lockdown error, retrying after a short sleep\n");
+        g_usleep (G_USEC_PER_SEC);
+        continue;
+      }
+
+    if (lerr != LOCKDOWN_E_PASSWORD_PROTECTED &&
+        lerr != LOCKDOWN_E_PAIRING_DIALOG_RESPONSE_PENDING)
       break;
 
     aborted = FALSE;
-    if (!message)
-      /* translators:
-       * %s is the device name. 'Try again' is the caption of the button
-       * shown in the dialog which is defined above. */
-      message = g_strdup_printf (_("The device “%s” is locked. Enter the passcode on the device and click “Try again”."), display_name);
+    if (lerr == LOCKDOWN_E_PASSWORD_PROTECTED)
+      {
+        /* translators:
+         * %s is the device name. 'Try again' is the caption of the button
+         * shown in the dialog which is defined above. */
+        message = g_strdup_printf (_("The device “%s” is locked. Enter the passcode on the device and click “Try again”."), display_name);
+      }
+    else if (lerr == LOCKDOWN_E_PAIRING_DIALOG_RESPONSE_PENDING)
+      {
+        /* translators:
+         * %s is the device name. 'Try again' is the caption of the button
+         * shown in the dialog which is defined above. */
+        message = g_strdup_printf (_("The device “%s” is not trusted yet. Select “Trust” on the device and click “Try again”."), display_name);
+      }
+    else
+      g_assert_not_reached ();
 
     ret = g_mount_source_ask_question (src,
                                        message,
                                        choices,
                                        &aborted,
                                        &choice);
+    g_free (message);
+
     if (!ret || aborted || (choice == CHOICE_CANCEL))
       break;
   } while (retries++ < 10);
@@ -614,15 +663,14 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
 
   g_free (display_name);
   display_name = NULL;
-  g_free (message);
 
-  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+  if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "initial paired client")))
     goto out_destroy_dev;
 
   switch (self->mode) {
     case ACCESS_MODE_AFC:
       lerr = lockdownd_start_service (lockdown_cli, self->service, &lockdown_service);
-      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "starting lockdownd")))
         {
           goto out_destroy_lockdown;
         }
@@ -634,7 +682,7 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
       break;
     case ACCESS_MODE_HOUSE_ARREST:
       lerr = lockdownd_start_service (lockdown_cli, "com.apple.mobile.installation_proxy", &lockdown_service);
-      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "starting install proxy")))
         {
           g_warning ("couldn't start inst proxy");
           goto out_destroy_lockdown;
@@ -646,7 +694,7 @@ g_vfs_backend_afc_mount (GVfsBackend *backend,
           goto out_destroy_lockdown;
         }
       lerr = lockdownd_start_service (lockdown_cli, "com.apple.springboardservices", &lockdown_service);
-      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job))))
+      if (G_UNLIKELY(g_vfs_backend_lockdownd_check (lerr, G_VFS_JOB(job), "starting install services")))
         {
           g_warning ("couldn't start SBServices proxy");
           goto out_destroy_lockdown;
@@ -761,8 +809,9 @@ is_regular (GVfsBackendAfc *backend,
   return result;
 }
 
-static void
+static gboolean
 g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
+                                 gboolean        last_try,
                                  const char     *id)
 {
   AppInfo *info;
@@ -772,27 +821,36 @@ g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
   afc_client_t afc;
   plist_t dict, error;
   lockdownd_error_t lerr;
+  gboolean retry = FALSE;
+
+  g_mutex_lock (&self->apps_lock);
 
   info = g_hash_table_lookup (self->apps, id);
 
   if (info == NULL ||
       info->afc_cli != NULL)
-    return;
+    goto out;
 
   /* Load house arrest and afc now! */
   lockdown_cli = NULL;
   if (lockdownd_client_new_with_handshake (self->dev, &lockdown_cli, "gvfsd-afc") != LOCKDOWN_E_SUCCESS)
     {
       g_warning ("Failed to get a lockdown to start house arrest for app %s", info->id);
-      return;
+      goto out;
     }
 
   lerr = lockdownd_start_service (lockdown_cli, "com.apple.mobile.house_arrest", &lockdown_service);
   if (lerr != LOCKDOWN_E_SUCCESS)
     {
+      if (lerr == LOCKDOWN_E_SERVICE_LIMIT && !last_try)
+        {
+          retry = TRUE;
+          g_debug ("Failed to start house arrest for app %s (%d)\n", info->id, lerr);
+        }
+      else
+        g_warning ("Failed to start house arrest for app %s (%d)", info->id, lerr);
       lockdownd_client_free (lockdown_cli);
-      g_warning ("Failed to start house arrest for app %s", info->id);
-      return;
+      goto out;
     }
 
   house_arrest = NULL;
@@ -802,7 +860,7 @@ g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
       g_warning ("Failed to start house arrest client for app %s", info->id);
       lockdownd_client_free (lockdown_cli);
       lockdownd_service_descriptor_free (lockdown_service);
-      return;
+      goto out;
     }
 
   lockdownd_service_descriptor_free (lockdown_service);
@@ -814,7 +872,7 @@ g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
       g_warning ("Failed to set up house arrest for app %s", info->id);
       house_arrest_client_free (house_arrest);
       lockdownd_client_free (lockdown_cli);
-      return;
+      goto out;
     }
   error = plist_dict_get_item (dict, "Error");
   if (error != NULL)
@@ -828,7 +886,7 @@ g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
 
       house_arrest_client_free (house_arrest);
       lockdownd_client_free (lockdown_cli);
-      return;
+      goto out;
     }
   plist_free (dict);
 
@@ -840,12 +898,109 @@ g_vfs_backend_setup_afc_for_app (GVfsBackendAfc *self,
     {
       g_warning ("Failed to set up afc client for app %s", info->id);
       house_arrest_client_free (house_arrest);
-
-      return;
+      goto out;
     }
 
   info->house_arrest = house_arrest;
   info->afc_cli = afc;
+
+out:
+  g_mutex_unlock (&self->apps_lock);
+  return !retry;
+}
+
+static FileHandle *
+g_vfs_backend_file_handle_new (GVfsBackendAfc *self,
+                               const char     *app)
+{
+  AppInfo *info;
+  FileHandle *handle;
+
+  handle = g_new0 (FileHandle, 1);
+
+  if (app == NULL)
+    return handle;
+
+  g_mutex_lock (&self->apps_lock);
+  info = g_hash_table_lookup (self->apps, app);
+  info->num_users++;
+  g_mutex_unlock (&self->apps_lock);
+
+  handle->app = g_strdup (app);
+
+  return handle;
+}
+
+static void
+g_vfs_backend_file_handle_free (GVfsBackendAfc *self,
+                                FileHandle     *fh)
+{
+  AppInfo *info;
+
+  if (fh == NULL)
+    return;
+
+  if (fh->app == NULL)
+    goto out;
+
+  g_mutex_lock (&self->apps_lock);
+  info = g_hash_table_lookup (self->apps, fh->app);
+  g_assert (info->num_users != 0);
+  info->num_users--;
+  g_mutex_unlock (&self->apps_lock);
+
+out:
+  if (self->connected)
+    afc_file_close (fh->afc_cli, fh->fd);
+  g_free (fh->app);
+  g_free (fh);
+}
+
+/* If we succeeded in removing access to at least one
+ * HouseArrest service, return TRUE */
+static gboolean
+g_vfs_backend_gc_house_arrest (GVfsBackendAfc *self,
+                               const char     *app)
+{
+  GList *apps, *l;
+  gboolean ret = FALSE;
+
+  g_mutex_lock (&self->apps_lock);
+
+  apps = g_hash_table_get_values (self->apps);
+  /* XXX: We might want to sort the apps so the
+   * oldest used gets cleaned up first */
+
+  for (l = apps; l != NULL; l = l->next)
+    {
+      AppInfo *info = l->data;
+
+      /* Don't close the same app we're trying to
+       * connect to the service, but return as it's
+       * already setup */
+      if (g_strcmp0 (info->id, app) == 0)
+        {
+          g_debug ("A HouseArrest service for '%s' is already setup\n", app);
+          ret = TRUE;
+          break;
+        }
+
+      if (info->afc_cli == NULL ||
+          info->num_users > 0)
+        continue;
+
+      g_clear_pointer (&info->afc_cli, afc_client_free);
+      g_clear_pointer (&info->house_arrest, house_arrest_client_free);
+
+      g_debug ("Managed to free HouseArrest service from '%s', for '%s'\n",
+               info->id, app);
+      ret = TRUE;
+      break;
+    }
+
+  g_mutex_unlock (&self->apps_lock);
+
+  return ret;
 }
 
 /* If force_afc_mount is TRUE, then we'll try to mount
@@ -854,12 +1009,12 @@ static char *
 g_vfs_backend_parse_house_arrest_path (GVfsBackendAfc *self,
                                        gboolean        force_afc_mount,
                                        const char     *path,
+                                       gboolean       *is_doc_root_ret,
                                        char          **new_path)
 {
   char **comps;
-  char *s;
   char *app;
-  gboolean setup_afc;
+  gboolean setup_afc, is_doc_root;
 
   if (path == NULL || *path == '\0' || g_str_equal (path, "/"))
     {
@@ -874,28 +1029,28 @@ g_vfs_backend_parse_house_arrest_path (GVfsBackendAfc *self,
 
   setup_afc = force_afc_mount;
   app = comps[0];
+  is_doc_root = (comps[0] != NULL && comps[1] == NULL);
+  if (is_doc_root_ret)
+    *is_doc_root_ret = is_doc_root;
 
   /* Replace the app path with "Documents" so the gvfs
    * path of afc://<uuid>/org.gnome.test/foo.txt should
    * correspond to Documents/foo.txt in the app's container */
   comps[0] = g_strdup ("Documents");
-  s = g_strjoinv ("/", comps);
-  if (*s == '\0')
-    {
-      g_free (s);
-      *new_path = g_strdup ("/");
-    }
-  else
-    {
-      *new_path = s;
-      setup_afc = TRUE;
-    }
+  *new_path = g_strjoinv ("/", comps);
+  if (is_doc_root)
+    setup_afc = TRUE;
   g_strfreev (comps);
 
   if (app != NULL &&
       setup_afc)
     {
-      g_vfs_backend_setup_afc_for_app (self, app);
+      if (!g_vfs_backend_setup_afc_for_app (self, FALSE, app))
+        {
+          g_debug ("Ran out of HouseArrest clients for app '%s', trying again\n", app);
+          g_vfs_backend_gc_house_arrest (self, app);
+          g_vfs_backend_setup_afc_for_app (self, TRUE, app);
+        }
     }
 
   return app;
@@ -912,21 +1067,19 @@ g_vfs_backend_afc_open_for_read (GVfsBackend *backend,
   char *new_path;
   afc_client_t afc_cli;
   FileHandle *handle;
+  char *app = NULL;
 
   self = G_VFS_BACKEND_AFC(backend);
   g_return_if_fail (self->connected);
 
   if (self->mode == ACCESS_MODE_HOUSE_ARREST)
     {
-      char *app;
       AppInfo *info;
+      gboolean is_doc_root;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &new_path);
-      if (app == NULL)
-        goto is_dir_bail;
-
-      if (g_str_equal (new_path, "/"))
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &is_doc_root, &new_path);
+      if (app == NULL || is_doc_root)
         goto is_dir_bail;
 
       info = g_hash_table_lookup (self->apps, app);
@@ -934,7 +1087,6 @@ g_vfs_backend_afc_open_for_read (GVfsBackend *backend,
         goto not_found_bail;
 
       afc_cli = info->afc_cli;
-      g_free (app);
     }
   else
     {
@@ -946,6 +1098,7 @@ g_vfs_backend_afc_open_for_read (GVfsBackend *backend,
     {
 is_dir_bail:
       g_free (new_path);
+      g_free (app);
       g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
                         G_IO_ERROR_IS_DIRECTORY,
                         _("Can't open directory"));
@@ -956,6 +1109,7 @@ is_dir_bail:
     {
 not_found_bail:
       g_free (new_path);
+      g_free (app);
       g_vfs_job_failed (G_VFS_JOB (job), G_IO_ERROR,
                         G_IO_ERROR_NOT_FOUND,
                         _("File doesn't exist"));
@@ -966,13 +1120,19 @@ not_found_bail:
                                                          new_path ? new_path : path, AFC_FOPEN_RDONLY, &fd),
                                           G_VFS_JOB(job))))
     {
+      g_free (new_path);
+      g_free (app);
       return;
     }
 
-  handle = g_new0 (FileHandle, 1);
+  g_free (new_path);
+
+  handle = g_vfs_backend_file_handle_new (self, app);
   handle->fd = fd;
   handle->afc_cli = afc_cli;
 
+  g_free (app);
+
   g_vfs_job_open_for_read_set_handle (job, handle);
   g_vfs_job_open_for_read_set_can_seek (job, TRUE);
   g_vfs_job_succeeded (G_VFS_JOB(job));
@@ -989,7 +1149,7 @@ g_vfs_backend_afc_create (GVfsBackend *backend,
 {
   uint64_t fd = 0;
   GVfsBackendAfc *self;
-  char *new_path, *app;
+  char *new_path, *app = NULL;
   afc_client_t afc_cli;
   FileHandle *fh;
 
@@ -1002,7 +1162,7 @@ g_vfs_backend_afc_create (GVfsBackend *backend,
     {
       AppInfo *info;
 
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, NULL, &new_path);
       if (app == NULL)
         {
           g_vfs_backend_afc_check (AFC_E_PERM_DENIED, G_VFS_JOB(job));
@@ -1011,11 +1171,12 @@ g_vfs_backend_afc_create (GVfsBackend *backend,
       info = g_hash_table_lookup (self->apps, app);
       if (info == NULL)
         {
+          g_free (new_path);
+          g_free (app);
           g_vfs_backend_afc_check (AFC_E_OBJECT_NOT_FOUND, G_VFS_JOB(job));
           return;
         }
       afc_cli = info->afc_cli;
-      g_free (app);
     }
   else
     {
@@ -1028,15 +1189,18 @@ g_vfs_backend_afc_create (GVfsBackend *backend,
                                           G_VFS_JOB(job))))
     {
       g_free (new_path);
+      g_free (app);
       return;
     }
 
   g_free (new_path);
 
-  fh = g_new0 (FileHandle, 1);
+  fh = g_vfs_backend_file_handle_new (self, app);
   fh->fd = fd;
   fh->afc_cli = afc_cli;
 
+  g_free (app);
+
   g_vfs_job_open_for_write_set_handle (job, fh);
   g_vfs_job_open_for_write_set_can_seek (job, TRUE);
   g_vfs_job_open_for_write_set_can_truncate (job, TRUE);
@@ -1055,7 +1219,7 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
   uint64_t fd = 0;
   uint64_t off = 0;
   GVfsBackendAfc *self;
-  char *new_path, *app;
+  char *new_path, *app = NULL;
   afc_client_t afc_cli;
   FileHandle *fh;
 
@@ -1068,7 +1232,7 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
     {
       AppInfo *info;
 
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, NULL, &new_path);
       if (app == NULL)
         {
           g_vfs_backend_afc_check (AFC_E_PERM_DENIED, G_VFS_JOB(job));
@@ -1077,11 +1241,12 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
       info = g_hash_table_lookup (self->apps, app);
       if (info == NULL)
         {
+          g_free (new_path);
+          g_free (app);
           g_vfs_backend_afc_check (AFC_E_OBJECT_NOT_FOUND, G_VFS_JOB(job));
           return;
         }
       afc_cli = info->afc_cli;
-      g_free (app);
     }
   else
     {
@@ -1094,6 +1259,7 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
                                           G_VFS_JOB(job))))
     {
       g_free (new_path);
+      g_free (app);
       return;
     }
 
@@ -1104,6 +1270,7 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
                                           G_VFS_JOB(job))))
     {
       afc_file_close (afc_cli, fd);
+      g_free (app);
       return;
     }
 
@@ -1112,13 +1279,16 @@ g_vfs_backend_afc_append_to (GVfsBackend *backend,
                                           G_VFS_JOB(job))))
     {
       afc_file_close (afc_cli, fd);
+      g_free (app);
       return;
     }
 
-  fh = g_new0 (FileHandle, 1);
+  fh = g_vfs_backend_file_handle_new (self, app);
   fh->fd = fd;
   fh->afc_cli = afc_cli;
 
+  g_free (app);
+
   g_vfs_job_open_for_write_set_handle (job, fh);
   g_vfs_job_open_for_write_set_can_seek (job, TRUE);
   g_vfs_job_open_for_write_set_can_truncate (job, TRUE);
@@ -1138,7 +1308,7 @@ g_vfs_backend_afc_replace (GVfsBackend *backend,
 {
   uint64_t fd = 0;
   GVfsBackendAfc *self;
-  char *new_path, *app;
+  char *new_path, *app = NULL;
   afc_client_t afc_cli;
   FileHandle *fh;
 
@@ -1161,7 +1331,7 @@ g_vfs_backend_afc_replace (GVfsBackend *backend,
     {
       AppInfo *info;
 
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, NULL, &new_path);
       if (app == NULL)
         {
           g_vfs_backend_afc_check (AFC_E_PERM_DENIED, G_VFS_JOB(job));
@@ -1170,11 +1340,12 @@ g_vfs_backend_afc_replace (GVfsBackend *backend,
       info = g_hash_table_lookup (self->apps, app);
       if (info == NULL)
         {
+          g_free (new_path);
+          g_free (app);
           g_vfs_backend_afc_check (AFC_E_OBJECT_NOT_FOUND, G_VFS_JOB(job));
           return;
         }
       afc_cli = info->afc_cli;
-      g_free (app);
     }
   else
     {
@@ -1187,15 +1358,18 @@ g_vfs_backend_afc_replace (GVfsBackend *backend,
                                           G_VFS_JOB(job))))
     {
       g_free (new_path);
+      g_free (app);
       return;
     }
 
   g_free (new_path);
 
-  fh = g_new0 (FileHandle, 1);
+  fh = g_vfs_backend_file_handle_new (self, app);
   fh->fd = fd;
   fh->afc_cli = afc_cli;
 
+  g_free (app);
+
   g_vfs_job_open_for_write_set_handle (job, fh);
   g_vfs_job_open_for_write_set_can_seek (job, TRUE);
   g_vfs_job_open_for_write_set_can_truncate (job, TRUE);
@@ -1217,10 +1391,7 @@ g_vfs_backend_afc_close_read (GVfsBackend *backend,
 
   self = G_VFS_BACKEND_AFC(backend);
 
-  if (self->connected)
-    afc_file_close (fh->afc_cli, fh->fd);
-
-  g_free (fh);
+  g_vfs_backend_file_handle_free (self, fh);
 
   g_vfs_job_succeeded (G_VFS_JOB(job));
 }
@@ -1237,10 +1408,7 @@ g_vfs_backend_afc_close_write (GVfsBackend *backend,
 
   self = G_VFS_BACKEND_AFC(backend);
 
-  if (self->connected)
-    afc_file_close(fh->afc_cli, fh->fd);
-
-  g_free (fh);
+  g_vfs_backend_file_handle_free (self, fh);
 
   g_vfs_job_succeeded (G_VFS_JOB(job));
 }
@@ -1301,7 +1469,7 @@ g_vfs_backend_afc_write (GVfsBackend *backend,
   g_vfs_job_succeeded (G_VFS_JOB(job));
 }
 
-static int
+static goffset
 g_vfs_backend_afc_seek (GVfsBackendAfc *self,
                         GVfsJob *job,
                         GVfsBackendHandle handle,
@@ -1310,12 +1478,13 @@ g_vfs_backend_afc_seek (GVfsBackendAfc *self,
 {
   int afc_seek_type;
   FileHandle *fh;
+  guint64 new_offset;
 
   if ((afc_seek_type = gvfs_seek_type_to_lseek (type)) == -1)
     {
       g_vfs_job_failed(job, G_IO_ERROR, G_IO_ERROR_INVALID_ARGUMENT,
                        _("Unsupported seek type"));
-      return 1;
+      return -1;
     }
 
   fh = (FileHandle *) handle;
@@ -1323,11 +1492,13 @@ g_vfs_backend_afc_seek (GVfsBackendAfc *self,
   if (G_UNLIKELY(g_vfs_backend_afc_check (afc_file_seek (fh->afc_cli,
                                                          fh->fd, offset, afc_seek_type),
                                           job)))
-    {
-      return 1;
-    }
+      return -1;
 
-  return 0;
+  if (G_UNLIKELY(g_vfs_backend_afc_check (afc_file_tell (fh->afc_cli, fh->fd, &new_offset),
+                                          job)))
+      return -1;
+
+  return new_offset;
 }
 
 static void
@@ -1338,15 +1509,17 @@ g_vfs_backend_afc_seek_on_read (GVfsBackend *backend,
                                 GSeekType type)
 {
   GVfsBackendAfc *self;
+  goffset new_offset;
 
   g_return_if_fail (handle != NULL);
 
   self = G_VFS_BACKEND_AFC(backend);
   g_return_if_fail (self->connected);
 
-  if (!g_vfs_backend_afc_seek (self, G_VFS_JOB(job), handle, offset, type))
+  new_offset = g_vfs_backend_afc_seek (self, G_VFS_JOB(job), handle, offset, type);
+  if (new_offset >= 0)
     {
-      g_vfs_job_seek_read_set_offset (job, offset);
+      g_vfs_job_seek_read_set_offset (job, new_offset);
       g_vfs_job_succeeded (G_VFS_JOB(job));
     }
 }
@@ -1359,15 +1532,17 @@ g_vfs_backend_afc_seek_on_write (GVfsBackend *backend,
                                  GSeekType type)
 {
   GVfsBackendAfc *self;
+  goffset new_offset;
 
   g_return_if_fail (handle != NULL);
 
   self = G_VFS_BACKEND_AFC(backend);
   g_return_if_fail (self->connected);
 
-  if (!g_vfs_backend_afc_seek (self, G_VFS_JOB(job), handle, offset, type))
+  new_offset = g_vfs_backend_afc_seek (self, G_VFS_JOB(job), handle, offset, type);
+  if (new_offset >= 0)
     {
-      g_vfs_job_seek_write_set_offset (job, offset);
+      g_vfs_job_seek_write_set_offset (job, new_offset);
       g_vfs_job_succeeded (G_VFS_JOB(job));
     }
 }
@@ -1842,7 +2017,7 @@ g_vfs_backend_afc_enumerate (GVfsBackend *backend,
           return;
         }
     }
-  else
+  else if (self->mode == ACCESS_MODE_HOUSE_ARREST)
     {
       char *app;
 
@@ -1855,7 +2030,7 @@ g_vfs_backend_afc_enumerate (GVfsBackend *backend,
         }
       g_mutex_unlock (&self->apps_lock);
 
-      app = g_vfs_backend_parse_house_arrest_path (self, TRUE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, TRUE, path, NULL, &new_path);
 
       if (app == NULL)
         {
@@ -1897,6 +2072,10 @@ g_vfs_backend_afc_enumerate (GVfsBackend *backend,
             }
         }
     }
+  else
+    {
+      g_assert_not_reached ();
+    }
 
   trailing_slash = g_str_has_suffix (new_path ? new_path : path, "/");
 
@@ -1962,9 +2141,10 @@ g_vfs_backend_afc_query_info (GVfsBackend *backend,
           return;
         }
     }
-  else
+  else if (self->mode == ACCESS_MODE_HOUSE_ARREST)
     {
       char *app;
+      gboolean is_doc_root;
 
       g_mutex_lock (&self->apps_lock);
       if (g_vfs_backend_load_apps (self) == FALSE)
@@ -1975,7 +2155,7 @@ g_vfs_backend_afc_query_info (GVfsBackend *backend,
         }
       g_mutex_unlock (&self->apps_lock);
 
-      app = g_vfs_backend_parse_house_arrest_path (self, TRUE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, TRUE, path, &is_doc_root, &new_path);
 
       if (app == NULL)
         {
@@ -1995,7 +2175,7 @@ g_vfs_backend_afc_query_info (GVfsBackend *backend,
               g_vfs_backend_afc_check (AFC_E_OBJECT_NOT_FOUND, G_VFS_JOB(job));
               return;
             }
-          if (g_str_equal (new_path, "/"))
+          if (is_doc_root)
             {
               g_free (new_path);
               g_vfs_backend_afc_set_info_from_app (self, info, app_info);
@@ -2010,6 +2190,10 @@ g_vfs_backend_afc_query_info (GVfsBackend *backend,
             }
         }
     }
+  else
+    {
+      g_assert_not_reached ();
+    }
 
   ptr = strrchr (new_path ? new_path : path, '/');
   if (ptr && ptr[1] != '\0')
@@ -2062,7 +2246,7 @@ g_vfs_backend_afc_query_fs_info (GVfsBackend *backend,
       AppInfo *info;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, NULL, &new_path);
       if (app == NULL)
         {
           g_vfs_backend_afc_check (AFC_E_OP_NOT_SUPPORTED, G_VFS_JOB(job));
@@ -2142,10 +2326,11 @@ g_vfs_backend_afc_set_display_name (GVfsBackend *backend,
     {
       char *app;
       AppInfo *info;
+      gboolean is_doc_root;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &afc_path);
-      if (app == NULL || g_str_equal (afc_path, "/"))
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &is_doc_root, &afc_path);
+      if (app == NULL || is_doc_root)
         {
           g_free (app);
           g_free (afc_path);
@@ -2184,9 +2369,15 @@ g_vfs_backend_afc_set_display_name (GVfsBackend *backend,
       return;
     }
 
-  g_vfs_job_set_display_name_set_new_path (job, new_path);
-  g_free (afc_path);
   g_free (new_path);
+  g_free (afc_path);
+
+  /* The new path, but in the original namespace */
+  dirname = g_path_get_dirname (filename);
+  new_path = g_build_filename (dirname, display_name, NULL);
+  g_vfs_job_set_display_name_set_new_path (job, new_path);
+  g_free (new_path);
+  g_free (dirname);
 
   g_vfs_job_succeeded (G_VFS_JOB(job));
 }
@@ -2221,10 +2412,11 @@ g_vfs_backend_afc_set_attribute (GVfsBackend *backend,
     {
       char *app;
       AppInfo *info;
+      gboolean is_doc_root;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &new_path);
-      if (app == NULL || g_str_equal (new_path, "/"))
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &is_doc_root, &new_path);
+      if (app == NULL || is_doc_root)
         {
           g_free (app);
           g_free (new_path);
@@ -2286,7 +2478,7 @@ g_vfs_backend_afc_make_directory (GVfsBackend *backend,
       AppInfo *info;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, &new_path);
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, path, NULL, &new_path);
       if (app == NULL)
         {
           g_free (app);
@@ -2398,17 +2590,18 @@ g_vfs_backend_afc_move (GVfsBackend *backend,
     {
       AppInfo *info;
       char *app_src, *app_dst;
+      gboolean is_doc_root;
 
-      app_src = g_vfs_backend_parse_house_arrest_path (self, FALSE, source, &new_src);
-      if (app_src == NULL || g_str_equal (new_src, "/"))
+      app_src = g_vfs_backend_parse_house_arrest_path (self, FALSE, source, &is_doc_root, &new_src);
+      if (app_src == NULL || is_doc_root)
         {
           g_free (app_src);
           g_free (new_src);
           g_vfs_backend_afc_check (AFC_E_PERM_DENIED, G_VFS_JOB(job));
           return;
         }
-      app_dst = g_vfs_backend_parse_house_arrest_path (self, FALSE, destination, &new_dst);
-      if (app_dst == NULL || g_str_equal (new_dst, "/"))
+      app_dst = g_vfs_backend_parse_house_arrest_path (self, FALSE, destination, &is_doc_root, &new_dst);
+      if (app_dst == NULL || is_doc_root)
         {
           g_free (app_src);
           g_free (new_src);
@@ -2430,6 +2623,9 @@ g_vfs_backend_afc_move (GVfsBackend *backend,
       info = g_hash_table_lookup (self->apps, app_src);
       if (info == NULL)
         {
+          g_free (app_src);
+          g_free (new_src);
+          g_free (new_dst);
           g_vfs_backend_afc_check (AFC_E_OBJECT_NOT_FOUND, G_VFS_JOB(job));
           return;
         }
@@ -2473,10 +2669,11 @@ g_vfs_backend_afc_delete (GVfsBackend *backend,
     {
       char *app;
       AppInfo *info;
+      gboolean is_doc_root;
 
       new_path = NULL;
-      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &new_path);
-      if (app == NULL || g_str_equal (new_path, "/"))
+      app = g_vfs_backend_parse_house_arrest_path (self, FALSE, filename, &is_doc_root, &new_path);
+      if (app == NULL || is_doc_root)
         {
           g_free (app);
           g_free (new_path);
