@@ -72,6 +72,12 @@
 #define PTP_ST_RemovableRAM                     0x0004
 
 
+/************************************************
+ * Constants
+ ************************************************/
+
+#define EVENT_POLL_PERIOD { 1, 0 }
+
 /************************************************
  * Private Types
  ************************************************/
@@ -96,6 +102,13 @@ typedef struct {
   uint32_t id;
 } CacheEntry;
 
+#if HAVE_LIBMTP_1_1_5
+typedef struct {
+  LIBMTP_event_t event;
+  uint32_t param1;
+} EventData;
+#endif
+
 
 /************************************************
  * Static prototypes
@@ -106,6 +119,11 @@ emit_delete_event (gpointer key,
                    gpointer value,
                    gpointer user_data);
 
+#if HAVE_LIBMTP_1_1_5
+static void
+handle_event (EventData *data, GVfsBackendMtp *backend);
+#endif
+
 
 /************************************************
  * Storage name helper
@@ -364,6 +382,11 @@ g_vfs_backend_mtp_init (GVfsBackendMtp *backend)
 
   backend->monitors = g_hash_table_new (NULL, NULL);
 
+#if HAVE_LIBMTP_1_1_5
+  backend->event_pool = g_thread_pool_new ((GFunc) handle_event,
+                                           backend, 1, FALSE, NULL);
+#endif
+
   g_debug ("(I) g_vfs_backend_mtp_init done.\n");
 }
 
@@ -384,6 +407,10 @@ g_vfs_backend_mtp_finalize (GObject *object)
 
   backend = G_VFS_BACKEND_MTP (object);
 
+#if HAVE_LIBMTP_1_1_5
+  g_thread_pool_free (backend->event_pool, TRUE, TRUE);
+#endif
+
   g_hash_table_foreach (backend->monitors, remove_monitor_weak_ref, backend->monitors);
   g_hash_table_unref (backend->monitors);
 
@@ -567,13 +594,65 @@ on_uevent (GUdevClient *client, gchar *action, GUdevDevice *device, gpointer use
     }
 
     op_backend->force_unmounted = TRUE;
+    g_atomic_int_set (&op_backend->unmount_started, TRUE);
     g_vfs_backend_force_unmount ((GVfsBackend*)op_backend);
   }
 
   g_debug ("(I) on_uevent done.\n");
 }
 
-#if HAVE_LIBMTP_1_1_5
+#if HAVE_LIBMTP_1_1_12
+static void
+check_event_cb(int ret, LIBMTP_event_t event, uint32_t param1, void *user_data)
+{
+  GVfsBackendMtp *backend = user_data;
+
+  g_debug ("(II) check_event_cb: %d, %d, %d\n", ret, event, param1);
+  backend->event_completed = TRUE;
+
+  if (ret != LIBMTP_HANDLER_RETURN_OK ||
+      g_atomic_int_get (&backend->unmount_started)) {
+    return;
+  }
+
+  EventData *data = g_new(EventData, 1);
+  data->event = event;
+  data->param1 = param1;
+  gboolean tret = g_thread_pool_push (backend->event_pool, data, NULL);
+  g_debug ("(II) check_event_cb push work to pool: %d\n", tret);
+}
+
+static gpointer
+check_event (gpointer user_data)
+{
+  GVfsBackendMtp *backend = user_data;
+
+  LIBMTP_event_t event;
+  while (!g_atomic_int_get (&backend->unmount_started)) {
+    int ret;
+    LIBMTP_mtpdevice_t *device = backend->device;
+    if (backend->event_completed) {
+      g_debug ("(I) check_event: Read event needs to be issued.\n");
+      ret = LIBMTP_Read_Event_Async (device, check_event_cb, backend);
+      if (ret != 0) {
+        g_debug ("(I) check_event: Read_Event_Async failed: %d\n", ret);
+        continue;
+      }
+      backend->event_completed = FALSE;
+    }
+    /*
+     * Return from polling periodically to check for unmount.
+     */
+    struct timeval tv = EVENT_POLL_PERIOD;
+    g_debug ("(I) check_event: Polling for events.\n");
+    ret = LIBMTP_Handle_Events_Timeout_Completed (&tv, &(backend->event_completed));
+    if (ret != 0) {
+      g_debug ("(I) check_event: polling returned error: %d\n", ret);
+    }
+  }
+  return NULL;
+}
+#elif HAVE_LIBMTP_1_1_5
 static gpointer
 check_event (gpointer user_data)
 {
@@ -600,15 +679,35 @@ check_event (gpointer user_data)
       return NULL;
     }
 
+    backend = g_weak_ref_get (event_ref);
+    if (backend && !g_atomic_int_get (&backend->unmount_started)) {
+      EventData *ed = g_new (EventData, 1);
+      ed->event = event;
+      ed->param1 = param1;
+      g_thread_pool_push (backend->event_pool, ed, NULL);
+      g_object_unref (backend);
+    }
+  }
+  return NULL;
+}
+#endif
+
+#if HAVE_LIBMTP_1_1_5
+void
+handle_event (EventData *ed, GVfsBackendMtp *backend)
+{
+  LIBMTP_event_t event = ed->event;
+  uint32_t param1 = ed->param1;
+  g_free (ed);
+
+  g_mutex_lock (&backend->mutex);
+  if (!g_atomic_int_get (&backend->unmount_started)) {
     switch (event) {
     case LIBMTP_EVENT_STORE_ADDED:
-      backend = g_weak_ref_get (event_ref);
-      if (backend && !g_atomic_int_get (&backend->unmount_started)) {
+      {
         LIBMTP_mtpdevice_t *device = backend->device;
         LIBMTP_devicestorage_t *storage;
 
-        g_mutex_lock (&backend->mutex);
-
         int ret = LIBMTP_Get_Storage (device, LIBMTP_STORAGE_SORTBY_NOTSORTED);
         if (ret != 0) {
           LIBMTP_Dump_Errorstack (device);
@@ -628,30 +727,14 @@ check_event (gpointer user_data)
             }
           }
         }
-
-        g_mutex_unlock (&backend->mutex);
-        g_object_unref (backend);
         break;
-      } else {
-        return NULL;
       }
 #if HAVE_LIBMTP_1_1_6
     case LIBMTP_EVENT_OBJECT_REMOVED:
-      backend = g_weak_ref_get (event_ref);
-      if (backend && !g_atomic_int_get (&backend->unmount_started)) {
-        g_mutex_lock (&backend->mutex);
         remove_cache_entry_by_id (G_VFS_BACKEND_MTP (backend), param1);
-        g_mutex_unlock (&backend->mutex);
-        g_object_unref (backend);
         break;
-      } else {
-        return NULL;
-      }
     case LIBMTP_EVENT_STORE_REMOVED:
-      backend = g_weak_ref_get (event_ref);
-      if (backend && !g_atomic_int_get (&backend->unmount_started)) {
-        g_mutex_lock (&backend->mutex);
-
+      {
         /* Clear the cache entries and emit delete event; first for all
            entries under the storage in question... */
         GHashTableIter iter;
@@ -671,18 +754,10 @@ check_event (gpointer user_data)
 
         /* ... and then for the storage itself */
         remove_cache_entry_by_id (G_VFS_BACKEND_MTP (backend), param1);
-
-        g_mutex_unlock (&backend->mutex);
-        g_object_unref (backend);
         break;
-      } else {
-        return NULL;
       }
     case LIBMTP_EVENT_OBJECT_ADDED:
-      backend = g_weak_ref_get (event_ref);
-      if (backend && !g_atomic_int_get (&backend->unmount_started)) {
-        g_mutex_lock (&backend->mutex);
-
+      {
         LIBMTP_file_t *object = LIBMTP_Get_Filemetadata(backend->device, param1);
         if (object) {
           /* Obtain parent's path by searching cache by object's parent
@@ -723,19 +798,14 @@ check_event (gpointer user_data)
 
           LIBMTP_destroy_file_t (object);
         }
-
-        g_mutex_unlock (&backend->mutex);
-        g_object_unref (backend);
         break;
-      } else {
-        return NULL;
       }
 #endif
     default:
       break;
     }
   }
-  return NULL;
+  g_mutex_unlock (&backend->mutex);
 }
 #endif
 
@@ -853,7 +923,10 @@ do_mount (GVfsBackend *backend,
     op_backend->hb_id =
       g_timeout_add_seconds (900, (GSourceFunc)mtp_heartbeat, op_backend);
 
-#if HAVE_LIBMTP_1_1_5
+#if HAVE_LIBMTP_1_1_12
+    op_backend->event_completed = TRUE;
+    op_backend->event_thread = g_thread_new ("events", check_event, backend);
+#elif HAVE_LIBMTP_1_1_5
     GWeakRef *event_ref = g_new0 (GWeakRef, 1);
     g_weak_ref_init (event_ref, backend);
     GThread *event_thread = g_thread_new ("events", check_event, event_ref);
@@ -884,6 +957,16 @@ do_unmount (GVfsBackend *backend, GVfsJobUnmount *job,
 
   g_atomic_int_set (&op_backend->unmount_started, TRUE);
 
+#ifdef HAVE_LIBMTP_1_1_12
+  /* Thread will terminate after flag is set. */
+  g_thread_join (op_backend->event_thread);
+#endif
+
+#if HAVE_LIBMTP_1_1_5
+  /* It's no longer safe to handle events. */
+  g_thread_pool_set_max_threads (op_backend->event_pool, 0, NULL);
+#endif
+
   /* Emit delete events to tell clients files are gone. */
   GHashTableIter iter;
   gpointer key, value;
